from xchainpy_client.models.tx_types import TX, TxHistoryParams, TxPage
from xchainpy_client.utxo_client import UTXOClient
from xchainpy_util.chain import Chain
from . import utils, crypto, haskoin_api
from . models.client_types import BitcoincashClientParams, BitcoincashTxParams
from bitcash import transaction, PrivateKey, PrivateKeyTestnet, network


class Client(UTXOClient):
    def __init__(self, params:BitcoincashClientParams):
        """
        :param params: params
        :type params: BitcoincashClientParams
        """
        UTXOClient.__init__(self, Chain.BitcoinCash, params)
        self.set_haskoin_url(params.haskoin_url)

    def set_haskoin_url(self, haskoin_url:str):
        """Set/Update the haskoin url

        :param haskoin_url: The new haskoin url
        :type haskoin_url: str
        """
        self.haskoin_url = haskoin_url
    
    def get_explorer_url(self) -> str:
        """Get explorer url
        :returns: the explorer url based on the network
        """
        return 'https://www.blockchain.com/bch-testnet' if self.network == 'testnet' else 'https://www.blockchain.com/bch'

    def get_explorer_address_url(self, address:str) -> str:
        """Get the explorer url for the given address

        :param address: address
        :type address: str
        :returns: The explorer url for the given address based on the network
        """
        return f'{self.get_explorer_url()}/address/{address}'

    def get_explorer_tx_url(self, tx_id:str) -> str:
        """Get the explorer url for the given transaction id

        :param tx_id: tx_id
        :type tx_id: str
        :returns: The explorer url for the given transaction id based on the network
        """
        return f'{self.get_explorer_url()}/tx/{tx_id}'


    def __get_private_key(self, index:int=0):
        """Get private key

        :param phrase: The phrase to be used for generating privkey
        :type phrase: str
        :returns: The privkey generated from the given phrase
        """
        try:
            derivation_path = self.root_derivation_paths.testnet if self.get_network() == 'testnet' else self.root_derivation_paths.mainnet
            derivation_path += str(index)
            privKey = crypto.mnemonic_to_private_key(self.phrase, derivation_path, self.network)
            return privKey
        except:
            raise Exception("Invalid Phrase")


    def get_address(self, index:int=0):
        """Get the current address

        :returns: the current address
        :raises: Raises if phrase has not been set before. A phrase is needed to create a wallet and to derive an address from it.
        """
        if index < 0:
            raise Exception('index must be greater than zero')
        
        if self.phrase:
            derivation_path = self.root_derivation_paths.testnet if self.get_network() == 'testnet' else self.root_derivation_paths.mainnet
            address = crypto.mnemonic_to_address(mnemonic=self.phrase, derivation_path=derivation_path + str(index), network=self.get_network())

            if not address:
                raise Exception('Address not defined')

            return address
        raise Exception('Phrase must be provided')

    def validate_address(self, address):
        """Validate the given address

        :param network: testnet or mainnet
        :type network: str
        :param address: address
        :type address: str
        :returns: True or False
        """
        return utils.validate_address(address)

    async def get_balance(self, address):
        """Get the BCH balance of a given address

        :param address: BCH address
        :type address: str
        :returns: The BCH balance of the address
        """
        balance = await utils.get_balance(self.haskoin_url, self.get_network(), address)
        return balance

    async def get_transactions(self, params:TxHistoryParams) -> TxPage:
        """Get transaction history of a given address with pagination options
        By default it will return the transaction history of the current wallet

        :param params: params
        :type params: TxHistoryParams
        :returns: The transaction history
        """
        try:
            account = await haskoin_api.get_account(self.haskoin_url, self.get_network(), params.address)
            if not account:
                    raise Exception("Invalid Address")

            offset = params.offset or 0
            limit = params.limit or 10
            transactions = await haskoin_api.get_transactions(self.haskoin_url, self.get_network(), params.address, offset, limit)
            if not transactions:
                    raise Exception("Transactions could not found for this address")

            total = account.txs
            txs = []
            for tx in transactions:
                tx = utils.parse_tx(tx)
                txs.append(tx)
 
            return TxPage(total=total, txs=txs)

        except Exception as err:
            raise Exception(str(err))

    async def get_transaction_data(self, tx_id:str) -> TX:
        """Get the transaction details of a given transaction id

        if you want to give a hash that is for mainnet and the current self.network is 'testnet',
        you should call self.set_network('mainnet') (and vice versa) and then call this method.

        :param tx_id: The transaction id
        :type tx_id: str
        :returns: The transaction details of the given transaction id
        """
        try:
            tx = await haskoin_api.get_transaction(self.haskoin_url, self.get_network(), tx_id)
            if not tx:
                raise Exception("Invalid tx_id")
            tx = utils.parse_tx(tx)
            return tx
        except Exception as err:
            raise Exception(str(err))

    async def _get_suggested_fee_rate(self):
        fee_rate = await haskoin_api.get_suggested_tx_fee()
        return fee_rate

    def _calc_fee(self, fee_rate, memo):
        fee = utils.calc_fee(fee_rate, memo)
        return fee

    async def transfer(self, params:BitcoincashTxParams):
        """Transfer BCH

        :param params: transfer params
        :type params: BitcoincashTxParams
        :returns: the transaction hash
        """
        try:
            if not params.fee_rate:
                fee_rates = await self.get_fee_rates()
                params.fee_rate = fee_rates.fast

            KeyClass = PrivateKeyTestnet if self.get_network() == "testnet" else PrivateKey
            key = KeyClass.from_hex(self.__get_private_key(params.wallet_index).Raw().ToHex())

            tx_hex = await utils.build_tx(self.haskoin_url, params.amount , params.recipient , params.memo , params.fee_rate , self.get_address(),self.get_network(), key)
            
            tx_id = await utils.broadcast_tx(self.get_network(), tx_hex)
            return tx_id
            
        except Exception as err:
            raise Exception(str(err))


import datetime
from xchainpy.xchainpy_bitcoincash.xchainpy_bitcoincash.models.api_types import TxUnspent
from cashaddress import convert
from bitcash import transaction
from xchainpy_util.asset import AssetBCH
from xchainpy_client.models.balance import Balance
from xchainpy_client.models import tx_types
from . import haskoin_api, bitcore_api



BCH_DECIMAL = 8
DEFAULT_SUGGESTED_TRANSACTION_FEE = 1

def parse_tx(tx):
    """Parse tx
    :param tx: The transaction to be parsed
    :type tx: str
    :returns: The transaction parsed from the binance tx
    """
    asset = AssetBCH
    tx_from = [tx_types.TxFrom(i['address'], i['value'] * 10 ** -8) for i in tx['inputs']]
    tx_to = [tx_types.TxTo(i['address'], i['value'] * 10 ** -8) for i in tx['outputs']]
    tx_date = datetime.datetime.fromtimestamp(tx['time'])
    tx_type = 'transfer'
    tx_hash = tx['txid']

    tx = tx_types.TX(asset, tx_from, tx_to, tx_date, tx_type, tx_hash)
    return tx

async def get_balance(haskoin_url:str, network:str, address:str):
    """Get the BCH balance of a given address

    :param haskoin_url: haskoin url
    :type haskoin_url: str
    :param address: BCH address
    :type address: str
    :param network: mainnet or testnet
    :type network: str
    :returns: The BCH balance of the address
    """
    try:
        account = await haskoin_api.get_account(haskoin_url, network, address)
        if not account:
                raise Exception("Invalid Address")
        
        balance = [Balance(AssetBCH, (account.confirmed + account.unconfirmed) * 10 ** -8)]
        return balance
    except Exception as err:
        raise Exception(str(err))

def calc_fee(fee_rate, memo=None, utxos=None):
    """Calculate fees based on fee rate and memo

    :param fee_rate: fee rates
    :type fee_rate: str
    :param memo: memo (optional)
    :type memo: str
    :param utxos: A list of utxos (optional)
    :type utxos: str
    :returns: The calculated fees based on fee rate and the memo
    """
    # key = PrivateKey()
    # random_address = key.address

    if utxos:
        utxos_len = len(utxos)
    else:
        utxos_len = 0

    # outputs = [(random_address, 0, 'bch')]
    # for i, output in enumerate(outputs):
    #     dest, amount, currency = output
    #     outputs[i] = (dest, transaction.currency_to_satoshi_cached(amount, currency))
    # num_outputs = len(outputs) + 1
    num_outputs = 2

    total_op_return_size = 0
    if memo:
        message_chunks = transaction.chunk_data(memo, transaction.MESSAGE_LIMIT)
        for message in message_chunks:
            total_op_return_size += transaction.get_op_return_size(message, custom_pushdata=False)

    calculated_fee = transaction.estimate_tx_fee(utxos_len, num_outputs, fee_rate, False, total_op_return_size)
    return calculated_fee

async def scan_UTXOs(haskoin_url, network, address):
    """Scan UTXOs from sochain

    :param haskoin_url: haskoin url
    :type haskoin_url: str
    :param network: testnet or mainnet
    :type network: str
    :param address: address
    :type address: str
    :returns: The UTXOs of the given address
    """
    unspents = await haskoin_api.get_unspent_transactions(haskoin_url, network, address)
    utxos = list(map(TxUnspent.bitcash_unspent_from_object, unspents))
    return utxos

def validate_address(address):
    return convert.is_valid(address)

async def build_tx(haskoin_url, amount, recipient, memo, fee_rate, sender, network, key):
    """Build transcation

    :param haskoin_url: haskoin url
    :type haskoin_url: str
    :param amount: amount of BCH to transfer
    :type amount: int
    :param recipient: destination address
    :type recipient: str
    :param memo: optional memo for transaction
    :type memo: str
    :param fee_rate: fee rates for transaction
    :type fee_rate: int
    :param sender: sender's address
    :type sender: str
    :param network: testnet or mainnet
    :type network: str
    :param client_url: The haskoin API url
    :type client_url: str
    :param key: bitcash object
    :type key: object
    :returns: transaction
    """
    try:
        if not validate_address(recipient):
            raise Exception('Invalid address')

        utxos = await scan_UTXOs(haskoin_url, network, sender)
        if len(utxos) == 0:
            raise Exception("No utxos to send")

        balance = await get_balance(haskoin_url, network, sender)

        if not balance:
            raise Exception("No BCH balance found")

        outputs = [(recipient, amount, 'bch')]

        tx = key.create_transaction(outputs=outputs, fee=int(fee_rate), message=memo, unspents=utxos, leftover=sender)
        return tx

    except Exception as err:
        raise Exception(str(err))

async def broadcast_tx(network, tx_hex):
    """Broadcast the transaction

    :param network: testnet or mainnet
    :type network: str
    :param tx_hex: tranaction hex
    :type tx_hex: str
    :returns: The transaction hash
    """
    return await bitcore_api.broadcast_tx(network, tx_hex)


from bitcash import Key, PrivateKeyTestnet
from bip_utils.bip.bip_keys import BipPrivateKey
from mnemonic import Mnemonic
from bip_utils import Bip32, P2WPKH, BitcoinCashConf
from bip_utils.conf import Bip32Conf

def mnemonic_to_seed(mnemonic, language="english"):
    """Convert mnemonic (phrase) to seed

    :param mnemonic: A phrase
    :type mnemonic: str
    :param language: language
    :type language: str
    :returns: seed
    """
    mnemo = Mnemonic(language)
    seed = mnemo.to_seed(mnemonic)
    return seed

def mnemonic_to_private_key(mnemonic:str, derivation_path:str, network:str, mnemonic_language:str='english') -> BipPrivateKey:
    """Convert mnemonic (phrase) to a private key

    :param mnemonic: A phrase
    :type mnemonic: str
    :param derivation_path: derivation_path. example: "m/84'/0'/0'/0/1"
    :type derivation_path: str
    :param network: for example 'mainnet' or 'testnet'
    :type network: str
    :returns: private_key
    """
    seed = mnemonic_to_seed(mnemonic=mnemonic, language=mnemonic_language)
    key_net_ver = Bip32Conf.KEY_NET_VER.Test() if network == 'testnet' else Bip32Conf.KEY_NET_VER.Main()
    bip32_ctx = Bip32.FromSeedAndPath(seed, derivation_path, key_net_ver)
    private_key = bip32_ctx.PrivateKey()

    return private_key

def mnemonic_to_address(mnemonic:str, derivation_path:str, network:str, mnemonic_language:str='english') -> str:
    """Convert a private key to a public key

    :param mnemonic: A phrase
    :type mnemonic: str
    :param network: for example 'mainnet' or 'testnet'
    :type network: str
    :param derivation_path: derivation_path
    :type derivation_path: str
    :param mnemonic_language: A language
    :type mnemonic_language: str
    :returns: address
    """
    seed = mnemonic_to_seed(mnemonic=mnemonic, language=mnemonic_language)
    key_net_ver = Bip32Conf.KEY_NET_VER.Test() if network == 'testnet' else Bip32Conf.KEY_NET_VER.Main()
    bip32_ctx = Bip32.FromSeedAndPath(seed, derivation_path, key_net_ver)
    private_key_hex = bip32_ctx.PrivateKey().Raw().ToHex()
    KeyClass = PrivateKeyTestnet if network == "testnet" else Key
    key = KeyClass.from_hex(private_key_hex)
    return key.address